/*
 * UNC01x firmware. (c) Unicontrollers, Ltd., 2010-2013
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file unc_prim.c
 * Primary functions of the device: what the device actually can do
 *
 * Written by Andrey V. Stolyarov & Svyatoslav Tolchenov
 *
 */

#include <stdint.h>
#include <stddef.h>
#include <avr/eeprom.h>

#include "ucd.h"
//#include "led.h"

#include "misc.h"
#include "print.h"
#include "factrydf.h"
#include "unc_prim.h"

#include "unc_1w.h"

#define MAX_DELAYED_ACTIONS 4

static uint32_t delayed_action_counters[MAX_DELAYED_ACTIONS];
static struct { uint8_t grp, act; } delayed_actions[MAX_DELAYED_ACTIONS];


static uint8_t input_timestamps[UNC_INPUTS_COUNT] =
    { 0, 0, 0, 0, 0, 0, 0, 0 };
uint8_t confirmed_input_state = 0;

struct unc001_global_configuration_str global_configuration;
struct unc001_global_state_str global_state;
struct unc001_extra_state_str_1 extra_state;

static uint8_t get_extension(uint8_t n)
{
#if 0  /* another impl. does precisely the same, but gains 8 bytes */
    return (global_configuration.hardware_config_bytes[0] >> (4*n)) & 0x0f;
#else
    uint8_t b = global_configuration.hardware_config_bytes[0];
    if(n)
        b >>= 4;
    return b & 0x0f;
#endif
}

uint8_t get_actual_input_state_byte()
{
    uint8_t res = 0;
    /*if(get_extension(1) == 0x04) {*/
    if(get_extension(1)) {
        uint8_t i20=(~IPORT2&(1<<IPORT2_IN0))>>IPORT2_IN0;
        uint8_t i21=(~IPORT2&(1<<IPORT2_IN1))>>IPORT2_IN1;
        uint8_t i22=(~IPORT2&(1<<IPORT2_IN2))>>IPORT2_IN2;
        uint8_t i23=(~IPORT2&(1<<IPORT2_IN3))>>IPORT2_IN3;
        res |= ((i20<<4) | (i21<<5) | (i22<<6) | (i23<<7));
    } else {
        res |= 0xf0;
    }
    /*if(get_extension(0) == 0x04) {*/
#if 0   /* 2 bytes gain */
    if(get_extension(0)) {
        res |= (~IPORT) & 0x0f;
    } else {
        res |= 0x0f;
    }
#else
    res |= get_extension(0) ? (~IPORT) & 0x0f : 0x0f;
#endif
    return res;
}

static void cancel_delayed_actions(uint8_t ac)
{
    uint8_t i;
    if(ac == 0xfe) /* official noop */
        return;
    for(i=0; i<MAX_DELAYED_ACTIONS; i++)
        if(ac == 0xff ||
            (delayed_actions[i].grp == 0x0e && delayed_actions[i].act == ac))
        {
            delayed_actions[i].grp = 0xff;
        }
}

static void init_software_state()
{
    zero_counters(NULL);
    global_state.state_bits_bytes[0] = 0;
    global_state.state_bits_bytes[1] = get_actual_input_state_byte();
    global_state.fsm_state = 0;
    global_state_drop_message();
    confirmed_input_state = get_actual_input_state_byte();
    cancel_delayed_actions(0xff /* all */);
}

static void init_timer()
{
    /* TCCR1B=0x05; // CLK/1024 */
    /* TCCR1B|=(1<<CS12)|(1<<CS10);  */
    /* TCCR1B|=(1<<CS12)|(1<<CS10);  */
    /* TCCR1B|=(1<<CS12); //CLK/256 */
    /* TCCR1B |= (1<<CS11); */ /* CLK/8 */

    /* CLK/8 */
    TCCR1B = 2;
    //OCR1AH=58;
    //OCR1AL=152;  /* 10 mS */
    //OCR1A = 15000;
    OCR1A = *((uint16_t*)&(global_configuration.hardware_config_bytes[2]));

#if 0
    TCCR1B = 4;  /* CLK/256 */
    OCR1AH=1;
    OCR1AL=213;  /* 10 mS */

    TCCR1B = 5;  /* CLK/1024 */
    OCR1AH=0;
    OCR1AL=117;  /* 10 mS */
#endif

    TIMSK|=(1<<OCIE1A); // Allow the interrupt upon match
}

void init_primary_hardware()
{
    eeprom_read_block(&global_configuration.hardware_config,
                      (unsigned char *)0 + 16, 48);

    if(get_extension(0) == 0x0f) {
        /* the first ext. is output */
        DDRC=0x0F;
        PORTC=0x00;
    } else {
        DDRC=0;
        PORTC=0x0F;
    }
    switch(get_extension(1)) {
    case 0x04:
        DDRB&=~0x3a;
        PORTB|=0x3a;
        break;
    default:
    case 0x0f:
        /* the second ext. is output */
        DDRB|=0x3a;
        PORTB&=~0x3a;
        break;
    }

    DDRD|=0xF0;
    PORTD&=0x0F;

    init_timer();
    init_software_state();
}

void restore_factory_defaults()
{
    uint8_t i;
    static const uint32_t def_react_reg = DEFAULT_REACTION_REGISTER;
    static const uint32_t def_user_cf = DEFAULT_USER_CONFIGURATION;
    for(i=0; i<UNC_INPUTS_COUNT; i++)
        memcpy_4b(global_configuration.react_bytes[i],
                  (uint8_t*)&def_react_reg);
    memcpy_4b(global_configuration.user_config_bytes, (uint8_t*)&def_user_cf);

    init_software_state();
}

void global_state_putchar(uint8_t ch)
{
    global_state.message_buffer[global_state.buf_end] = ch;
    ++global_state.buf_end;
    global_state.buf_end &= 0x1f;
    if(global_state.buf_end == global_state.buf_start) {
        ++global_state.buf_start;
        global_state.buf_start &= 0x1f;
#if 0    /* this costs 12 bytes... too much */
        global_state.message_buffer[global_state.buf_start] = '>';
#endif
    }
}

void global_state_drop_message()
{
    global_state.buf_start = 0;
    global_state.buf_end = 0;
}

void sync_global_state()
{
    uint8_t module_mask = 
        ((get_extension(0) == 0x04) ? 0x0f : 0x00) |
        ((get_extension(1) == 0x04) ? 0xf0 : 0x00);

    global_state.state_bits_bytes[0] =
        ((OPORT&(1<<OUBIT_A))>>OUBIT_A)|
        ((OPORT&(1<<OUBIT_B))>>OUBIT_B<<1)|
        ((OPORT&(1<<OUBIT_C))>>OUBIT_C<<2)|
        ((OPORT&(1<<OUBIT_D))>>OUBIT_D<<3);

    global_state.state_bits_bytes[1] =
        (confirmed_input_state & module_mask) |
        (~get_actual_input_state_byte() & ~module_mask);
}

void save_registers_to_eeprom()
{
    eeprom_write_block(&global_configuration.hardware_config,
                       (unsigned char *)0 + 16, 48);
}

void zero_counters(uint16_t *save)
{
    uint8_t i;
    for(i=0; i<UNC_INPUTS_COUNT; i++) {
        /* if the counter is incremented both by pressing and releasing the
           button, *AND* the "keep parity" flag is set, then the counter's
           parity should reflect the current state of the button
         */
        uint8_t should_be_one =
            (
                ((global_configuration.react_bytes[i][3] & 0x70) == 0x70) &&
                (confirmed_input_state & (1<<i))
            ) ? 1 : 0; 
#if 0
        /* this proved to be not a good idea: it is better to give
           the requester the opportunity to check (by the counters'
           value) whether the button is pressed now or not
         */
        if(save) {
            save[i] = global_state.counters[i] - should_be_one;
        }
#else
        if(save) {
            save[i] = global_state.counters[i];
        }
#endif
        global_state.counters[i] = should_be_one;
    }
}

static void perform_extended_action(uint8_t regnum)
{
    uint8_t reg[8];
    int8_t i;
    uint8_t n;
    union uuu {
        uint8_t b[2];
        uint16_t a;
    } ua;
    static const uint32_t mults[] = { 1, 100, 6000, 360000 };
    static uint8_t recursion_protector = 0;

    if(recursion_protector > 6) return;

    if(regnum >= 48) /* 0xfe or 0xff mean noop, others are unhandled error */
        return;
    eeprom_read_block(reg, (uint8_t *)128 + regnum * 8, 8);

    ++recursion_protector;

    if((reg[7] & 0xf0) == 0xe0) {
        for(i = 14; i >= 0; i--) {
            if(i % 3 == 2)
                ua.a = 0;
            ua.a <<= 4;
            ua.a |= (reg[i>>1] >> ( (i & 1) ? 4 : 0 )) & 0x0f;
            if(i % 3 == 0) {
#if 0
                print_char('&');
                print_hex_byte(ua.b[1]);
                print_hex_byte(ua.b[0]);
                print_char('.');
#endif
                perform_action(ua.b[1], ua.b[0]);
            }
        }
    }
    else
    switch(reg[7]) {
    case 0xf0:
        perform_action(reg[5], reg[4]);
        /* now set up the delayed action, if any */
        /* this must be done AFTER the immediate action, as 
           the immediate one could be "cancel all delayed" */
        if(reg[1] != 0xff) { /* don't waste slots for noOps */
            if(reg[3]==0) { /* it is not really "delayed", just execute now */
                perform_action(reg[1], reg[0]);
            } else {
                for(i=0; i<MAX_DELAYED_ACTIONS; i++)
                    if(delayed_actions[i].grp == 0xff) {
                        /* here we go! */
                        delayed_actions[i].grp = reg[1];
                        delayed_actions[i].act = reg[0];
                        delayed_action_counters[i] = reg[3]*mults[reg[2]&0x3];
                        break;
                    }
            }
        }
        break;
    case 0xf4:
    case 0xf5:
    case 0xf6:
    case 0xf7:
        n = (global_state.fsm_state >> ((reg[7] & 0x03) * 4)) & 0x0f;
        if(n <= ((reg[6] & 0xf0) >> 4))
            perform_action(reg[5], reg[4]);
        else
        if(n <= (reg[6] & 0x0f))
            perform_action(reg[3], reg[2]);
        else
            perform_action(reg[1], reg[0]);
    }
    --recursion_protector;
}

static void process_input_change(uint8_t i)
{
    static char in_msg[] = "L+0\n";
    uint8_t now_on = confirmed_input_state & (1<<i);
    perform_action(
        global_configuration.react_bytes[i][2] & 0x0f, 
        global_configuration.react_bytes[i][now_on ? 0 : 1]
    );
    if(global_configuration.react_bytes[i][3] & (now_on ? 0x20 : 0x40))
        global_state.counters[i]++;
    if(global_configuration.react_bytes[i][3] & (now_on ? 0x04 : 0x08)) {
        /* in_msg[0] = 'L'; */
        in_msg[1] = now_on ? '+' : '-';
        in_msg[2] = '0' + i;
        /* in_msg[3] = '\n'; */
        print_str(in_msg);
    }
} 

void perform_action(uint8_t grp, uint8_t action)
{
    static volatile uint8_t* ports[3] = { &OPORT, &OPORT1, &OPORT2 };
    static const uint8_t bits[3][4] = {
        { 1<<OUBIT_A,  1<<OUBIT_B,  1<<OUBIT_C,  1<<OUBIT_D  },
        { 1<<OUBIT1_A, 1<<OUBIT1_B, 1<<OUBIT1_C, 1<<OUBIT1_D },
        { 1<<OUBIT2_A, 1<<OUBIT2_B, 1<<OUBIT2_C, 1<<OUBIT2_D }
    };
    uint8_t i;
    uint16_t n;
    uint8_t b;

    switch(grp) {
    case 0:
    case 1:
    case 2:
        for(i = 0; i < 4; i++) {
            b = bits[grp][i];
            switch((action >> (2 * i)) & 3 /* least 2 bits */) {
                case 0:   /* 00: no-op */
                    break;
                case 1:   /* 01: on */
                    *ports[grp] |= b;
                    break;
                case 2:   /* 10: off */
                    *ports[grp] &= ~b;
                    break;
                case 3:   /* 11: switch */
                    *ports[grp] ^= b;
                    break;
            }
        }
        break;
    case 0x09:
        perform_1wire_action(action);
        break;
    case 0x0a:  /* was 0xe0 */
        /* i = ((action & 0x30) >> 4) * 4; */ /* rmmbr: each FSM is 4 bits */
        i = (action >> 2) & 0x0C; 
            /* i is now the FSM number SHL'ed by 2, that is, the bit pos */
        n = action & 0x0f;    /* n is the new value for the fsm */
        n <<= i;              /* now n is the value shifted to the pos. */
        global_state.fsm_state =
            (global_state.fsm_state & ~(0x000f << i)) | n;
        break;
    case 0x0b: /* bring */
        for(i = 0; i < 8; i++)
            if(action & (1 << i))
                process_input_change(i);
        break;
    case 0x0c: /* was 0xfe */ /* cancel delays */
        cancel_delayed_actions(action);
        break;
    case 0x0d:  /* was 0xfd */ /* output a message */
        print_str("A:");
        print_hex_byte(action);
        print_char('\n');
        break;
    case 0x0e:  /* was 0xf0 */
        perform_extended_action(action);
        break;
    case 0x0f: /* no operation */
    default:   /* actually an error */
        break;
    } 
}

void do_input_confirmation()
{
    uint8_t input_diff =
        get_actual_input_state_byte() ^ confirmed_input_state;
    uint8_t i;
    for(i = (get_extension(0) == 0x04) ? 0 : 4;
        i < UNC_INPUTS_COUNT; i++)
    {
#if 0
        if(global_configuration.react_bytes[i][3] & 0x80) /* analog mode */
            continue;
#endif
        if(i>=4 && (get_extension(1) != 0x04))
            break;
        if(input_diff & (1<<i)) {
            if(!--input_timestamps[i])  
            {
                /* yes it has really changed */
                confirmed_input_state ^= (1<<i);
                /* now handle the I-register */
                process_input_change(i);
            }
        } else { /* no difference */
            input_timestamps[i] =
                global_configuration.user_config_bytes[0] + 1;
        }
    }
}

void process_delayed_actions()
{
    uint8_t i, grp;
    for(i=0; i<MAX_DELAYED_ACTIONS; i++)
        if((grp = delayed_actions[i].grp) != 0xff) {
            if(--(delayed_action_counters[i]) == 0) {
                delayed_actions[i].grp = 0xff;
                perform_action(grp, delayed_actions[i].act);
            }
        }
}



void startup_hook()
{
    uint8_t ac = global_configuration.user_config_bytes[3] & 0x3f;
    if(!ac || ac > 48)
        return;
    if(ac == 48)
        ac = 0;
    perform_extended_action(ac);
}
