/*
 * UNC0xx GUI-based software. (c) Unicontrollers, Ltd., 2012, 2013
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file devlist.cpp
 * Implementation of the device communication subsystem (manages
 * the list of connected devices and allows to communicate with them)
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <stdio.h> // for sprintf
#include <uncusb.h>
#include <unc_data.h>

#include "config.hpp"

#include "devlist.hpp"

UncDeviceCollection* CreateDeviceCollection()
{
    return new UncDevList();
}


UncDevice::UncDevice(int a_id, void *a_dev_handle)
    : id(a_id), dev_handle(a_dev_handle), last_state(0), minc(0), maxc(0)
{
    counters = new int[UNC_INPUTS_COUNT];
    for(int i=0; i<UNC_INPUTS_COUNT; i++)
        counters[i] = 0;
    unc001_get_counter_range(dev_handle, &minc, &maxc);
}

UncDevice::~UncDevice()
{
    delete[] counters;
    if(dev_handle)
        unc001_close(dev_handle);
}

bool UncDevice::SetLine(int n, bool on)
{
    int action = (on ? 1 : 2) << ((n%4)*2);
    int group = n / 4;
    action |= (group << 8);
    if(!dev_handle) return false;
    return -1 != unc001_perform_action(dev_handle, action);
}

void UncDevice::ZeroCounters()
{
    if(!dev_handle) return;
    int c[UNC_INPUTS_COUNT];
    unc001_zero_query_counters(dev_handle, c, UNC_INPUTS_COUNT);
}

bool UncDevice::IsAlive() const
{
    return dev_handle != 0;
}

static char *process_1w_item(char *bp, const uint8_t *id, int temp)
{
    int j, grad, migrad;
    char tsign = '+';
    if(!id[0])
        return bp;
    bp += sprintf(bp, "1w: ");
    for(j=7; j>=0; j--)
        bp += sprintf(bp, "%02x", id[j]);
    switch(temp) {
    case (short)0x8000: /* no temperature here */
        *bp++ = '\n';
        break;
    case 0x7fff: /* crc error */
        bp += sprintf(bp, "    temperature: CRC error\n");
        break;
    default:
        grad = temp >> 4;
        migrad = (temp & 0x0f) * 625;
        if(temp < 0) {
            tsign = '-';
            if(migrad > 0) {
                migrad = 10000 - migrad;
                grad = -grad - 1;
            } else {
                grad = -grad;
            }
        }
        bp += sprintf(bp, "    temp: (0x%04hx) %c%d.%04d C\n",
                               temp, tsign, grad, migrad);
    }
    return bp;
}

void UncDevice::SyncState()
{
    struct unc001_global_state_str st;
    if(!dev_handle) return;
    if(-1 == unc001_get_fullstate(dev_handle, &st)) {
        unc001_close(dev_handle);
        dev_handle = 0;
        the_set->Detach();
        the_set->UpdateConnectedState();
        return;
    }
    if(st.state_bits != last_state) {
        if(the_set) {
            for(int r = 0; r<16; r++) {
                uint16_t m = 1 << r;
                if((st.state_bits & m) != (last_state & m)) {
                    // smth. changed
                    if(r<=3) {  /* basic output group */
                        the_set->SetOutput(r, st.state_bits & m);
                    } else
                    if(r<=7) { /* reserved bits, ignore them */
                    } else
                    if(r<=11) { /* for unc002, could be either in or out */
                        if(minc == 0 && maxc > 0) {  /* in */
                            the_set->SetInput(r-8, st.state_bits & m);
                        } else { /* out */
                            the_set->SetOutput(r-4, st.state_bits & m);
                        }
                    } else {
                        if(maxc > 3) {  /* in */
                            int lnum = (minc==0) ? r-8 : r-12;
                            the_set->SetInput(lnum, st.state_bits & m);
                        } else { /* out */
                            int lnum = (minc==0) ? r-4 : r-8;
                            the_set->SetOutput(lnum, st.state_bits & m);
                        }
                    }
                }
            } 
        }
        last_state = st.state_bits;
    }
    for(int i = minc; i<=maxc; i++) {
        if(st.counters[i] != counters[i]) {
            counters[i] = st.counters[i];
            if(the_set) {
                int lnum = (minc==0) ? i : i-4; 
                the_set->SetCounter(lnum, counters[i]);
            }
        }
    }
//////////////
    unc001_extra_state_str_common extra;
    int r = unc001_get_extrastate(dev_handle, &extra);
    if(r == -1) {
        the_set->SetStatusBar("Error getting extra state");
        return;
    }
    static char statusbuf[1024];
    char *bp = statusbuf;
    *bp = 0;
    if(extra.page_version == 0) {
        unc001_extra_state_str_0 *ex = (unc001_extra_state_str_0*)&extra; 
        for(int i=0; i<4; i++) {
            bp = process_1w_item(bp, ex->ow_ids[i], ex->ow_temperature[i]);
        }
    } else
    if(extra.page_version == 1) {
        unc001_extra_state_str_1 *ex = (unc001_extra_state_str_1*)&extra; 
        for(int i=0; i<6; i++) {
            bp = process_1w_item(bp, ex->ow_ids[i], ex->ow_temperature[i]);
        }
    } else {
        the_set->SetStatusBar("Extra state version unknown");
        return;
    }
    the_set->SetStatusBar(statusbuf);
}

bool UncDevice::GetRegisters(unsigned long *hw,
                             unsigned long *user,
                             unsigned long reacts[8])
{
    if(!dev_handle) return false;
    struct unc001_global_configuration_str conf;
    int r = unc001_get_configuration(dev_handle, &conf);
    if(r == -1) return false;
    *hw = conf.hardware_config;
    *user = conf.user_config;
    for(int i = 0; i < UNC_INPUTS_COUNT; i++) reacts[i] = conf.reactions[i];
    return true;
}

bool UncDevice::SetUserReg(unsigned long user)
{
    if(!dev_handle) return false;
    return unc001_set_configuration(dev_handle, user) != -1;
}

bool UncDevice::SetHardwareConf(unsigned long hw)
{
    if(!dev_handle) return false;
    return unc001_set_hardware_conf(dev_handle, hw) != -1;
}

bool UncDevice::SetReactionReg(int line, unsigned long reg)
{
    if(!dev_handle) return false;
    return unc001_set_reactions(dev_handle, line, reg) != -1;
}

bool UncDevice::SaveToEeprom()
{
    if(!dev_handle) return false;
    return unc001_save_to_eeprom(dev_handle) != -1;
}

bool UncDevice::RestoreFromEeprom()
{
    if(!dev_handle) return false;
    return unc001_restore_from_eeprom(dev_handle) != -1;
}

bool UncDevice::RestoreFactoryDefaults()
{
    if(!dev_handle) return false;
    return unc001_restore_factory(dev_handle) != -1;
}

bool UncDevice::SetId(int newid)
{
    if(!dev_handle) return false;
    if(unc001_set_new_id(dev_handle, newid) == -1)
        return false;
    id = newid;
    if(the_set) {
        the_set->Detach();
        the_set->UpdateConnectedState();
        the_set = 0;
    }
    return true;
}

void UncDevice::CheckHardwareConfiguration()
{
    if(!dev_handle) return;
    struct unc001_global_configuration_str conf;
    int r = unc001_get_configuration(dev_handle, &conf);
    if(r != -1) { // XXXX actually we need some error handling here...
        DeviceConfig *dconf = GetDeviceConfig(id);
        int m0 = conf.hardware_config_bytes[0] & 0xf;
        int m1 = (conf.hardware_config_bytes[0] & 0xf0) >> 4;
        if(dconf->GetModuleType(0) != m0)
            dconf->SetModuleType(0, (line_type)m0, true); 
        if(dconf->GetModuleType(1) != m1)
            dconf->SetModuleType(1, (line_type)m1, true); 
    }
}

void UncDevList::Rescan(bool showhidden)
{
    static bool usb_initialized = false;
    if(!usb_initialized) {
        usb_initialized = true;
        unc001_usb_init();
    }

    Clear(); // start from the blank!

    unc001_usb_rescan();
    unc001_item *list = unc001_scan(0 /* dev_id: scan for all */);
    for(unc001_item *t = list; t; t = t->next) {
        if(t->version==0 || !t->dev_hdl) {
            // deprecated version of the firmware
            continue; // XXX we should display this fact somehow...
        }
        int id = t->sernum[2];

        DeviceConfig *dconf = GetDeviceConfig(id);
        device_visibility dv = dconf->GetVisibility();
        if(dv == devv_ignored) continue;

        UncDevice *d = new UncDevice(id, t->dev_hdl);
        Add(d);
        t->dev_hdl = 0; // not to be closed on the list disposal

        if(dv == devv_hidden) {
            if(!showhidden)
                continue;
            dconf->SetVisibility(devv_default);
            dconf->Save();
        }

        d->CheckHardwareConfiguration();

        // now make sure it is displayed
        UncCtrlsetInterface *ctrl = the_ctrlsetcoll->Provide(id);
        ctrl->Attach(d);
        d->Attach(ctrl);
        ctrl->UpdateConnectedState();
    }
    unc001_destroy_list(list);

    // Now show the devices that are absent but configured as always shown
    ForciblyShownDevices fsd;
    for(int i = 0; i<fsd.count; i++)
        the_ctrlsetcoll->Provide(fsd.array[i])->UpdateConnectedState();
}

void UncDevList::Poll()
{
    Iterator iter(this);
    UncDevice *p;
    bool all_alive = true;
    while((p = static_cast<UncDevice*>(iter.Next()))) {
        if(p->IsDisplayed())
            p->SyncState();
        all_alive = all_alive && p->IsAlive();
    }
    if(!all_alive) {
        static unsigned char c = 20; // XXX should be ajustable
        if(!--c) { Rescan(false); c = 20; }
    }
}

UncDeviceInterface *UncDevList::Find(int id)
{
    Iterator iter(this);
    UncDevice *p;
    while((p = static_cast<UncDevice*>(iter.Next()))) {
        if(p->Id() == id)
            return p;
    }
    return 0;
}
